import pygame
from pygame.locals import *

import customsprites
import resource
import status
from helper import *

def init():
	pygame.init()
	global screen
	screen = Screen(pygame.display.set_mode((1024,768), pygame.HWSURFACE | pygame.DOUBLEBUF))
	#screen = Screen(pygame.display.set_mode((1024,768), pygame.FULLSCREEN | pygame.HWSURFACE | pygame.DOUBLEBUF))

def get_screen():
	global screen
	return screen

def get_map():
	return get_screen().map

class Screen:
	def __init__(self, scr):
		self.display_surface = scr
		self.memory_surface = pygame.Surface((1024,768), 0, scr)
		self.memory_surface.fill(pygame.color.Color('black'))
		self.background = pygame.Surface((1024,768), 0, scr)
		self.background.fill(pygame.color.Color('black'))
		self.sprites = pygame.sprite.LayeredUpdates()
		self.status = pygame.Surface((1024,200), 0, scr)
		
		self.tile_types = {1: 'GreenTile', 2: 'GardenTile', 3: 'GrayTile', 4: 'DarkenTile', 5:'BrownTile', 32:'HighlightTile' }
		
		self.mapoffset = (32, 260)
		
		self.cursor = customsprites.Cursor()
		self.sprites.add(self.cursor, layer=9001)
		self.highlight = None
	
	def reset(self):
		self.memory_surface.fill(pygame.color.Color('black'))
		self.background.fill(pygame.color.Color('black'))
		
		self.sprites = pygame.sprite.LayeredUpdates()
		self.cursor = customsprites.Cursor()
		self.sprites.add(self.cursor, layer=9001)
		
		self.highlight = None
	
	def set_map(self, new_map):
		self.map = new_map

	def add_sprite(self, sprite, t_pos):
		self.sprites.add(sprite)
		sprite.set_location(self.tile_to_screen(t_pos))

	def del_sprite(self, sprite):
		self.sprites.remove(sprite)

	def mouse_only(self):
		self.display_surface.blit(self.cursor.image, self.cursor.rect)
		
		pygame.display.flip()

	def update(self, tick):		
		# update sprites as needed
		self.sprites.update(tick)
		
		
		# draw status screen to memory surface
		self.memory_surface.blit(self.status, pygame.Rect(0, 0, 0, 0)) 
		
		# draw sprites to memory surface
		self.sprites.draw(self.memory_surface)
		
		# blit memory subsurface to display
		self.display_surface.blit(self.memory_surface, (0, 0))
		for x in self.map.mobs + self.map.terrains:
			if x.damage > 0 and x.vitality > 0:
				pos = x.sprite.rect.move(0, x.sprite.image.get_height() - 5).topleft
				long = TILE_PIXEL_WIDTH * min(1.0, max(0.0, (1 - (x.damage / float(x.vitality)))))
				pygame.draw.rect(self.display_surface, pygame.Color(120,0,0), pygame.Rect(pos, (TILE_PIXEL_WIDTH, 4)))
				pygame.draw.rect(self.display_surface, pygame.Color(0,255,0), pygame.Rect(pos, (long, 3)))
		
		
		pygame.display.flip()
		
		# clear sprites
		self.sprites.clear(self.memory_surface, self.background)


	def cursor_moved(self):
		self.cursor.update_position()
		self.highlight_under_cursor()

	def mouse_tile(self):
		return self.highlight

	def object_clicked(self):
		p = pygame.mouse.get_pos()
		obj = None
		for x in self.map.mobs + self.map.terrains:
			if pygame.Rect(x.sprite.rect.topleft, x.sprite.image.get_size()).collidepoint(p):
				obj = x
		#~ if obj != None:
			#~ print obj.name
		return obj

	def highlight_under_cursor(self):
		if self.highlight != None:
			self.map.render_tile(self.highlight)
		pos = self.screen_to_tile(pygame.mouse.get_pos())
		if self.map.valid_tile(pos):
			self.draw_tile(32, self.map.tile_to_pixel(pos))
			self.highlight = pos
		else:
			self.highlight = None

	def update_status_screen(self, summon):
		# show info for selected mob/terrain
		# show general info (mana, units lost, invaders remaining, &c.)
		self.status.fill(pygame.Color('#404040'))
		font = resource.get_font('bluehigh', 36)
		smfont = resource.get_font('bluehigh', 18)
		
		pygame.draw.rect(self.status, pygame.Color('black'), (10,110,200,40))
		self.status.blit(font.render("Mana: " + str(status.get_mana()), True, pygame.Color('blue')), (15, 115))
		
		pygame.draw.rect(self.status, pygame.Color('black'), (250,110,250,40))
		self.status.blit(font.render("Invaders: " + str(self.map.invaders_left), True, pygame.Color('red')), (255, 115))
		
		pygame.draw.rect(self.status, pygame.Color('black'), (10,10,600,70))
		if summon != None:
			mana_color = pygame.Color('blue')
			if summon.cost > status.get_mana():
				mana_color = pygame.Color('red')
			self.status.blit(font.render(str(summon.cost) + ' Mana to summon', True, mana_color), (65, 15))
			self.status.blit(font.render(summon.name, True, pygame.Color('gray')), (65, 45))
			self.status.blit(summon.sprite.image, (15, 70 - summon.sprite.image.get_height()))
			
			self.status.blit(smfont.render('Attack:', True, pygame.Color('darkgray')), (400, 15))
			self.status.blit(smfont.render('Vitality:', True, pygame.Color('darkgray')), (400, 35))
			self.status.blit(smfont.render('Speed:', True, pygame.Color('darkgray')), (400, 55))
			self.status.blit(smfont.render(str(summon.attack), True, pygame.Color('gray')), (465, 15))
			self.status.blit(smfont.render(str(summon.vitality), True, pygame.Color('gray')), (465, 35))
			self.status.blit(smfont.render(str(int(summon.sprite.speed * 10 / summon.sprite.cycle)), True, pygame.Color('gray')), (465, 55))
		else:
			self.status.blit(font.render('No summoning selected', True, pygame.Color('gray')), (65, 45))

	def draw_tile(self, tile, pos):
		t = resource.get_image(self.tile_types[tile])
		self.background.blit(t, pos.move(self.mapoffset))
		self.memory_surface.blit(t, pos.move(self.mapoffset))
	
	def tile_to_screen(self, t_pos):
		return self.pixel_to_screen(self.map.tile_to_pixel(t_pos))
	
	def pixel_to_screen(self, p_pos):
		return tup2rect(p_pos).move(self.mapoffset)
	
	def screen_to_pixel(self, s_pos):
		return tup2rect(s_pos).move(-self.mapoffset[0], -self.mapoffset[1])
	
	def screen_to_tile(self, s_pos):
		return self.map.pixel_to_tile(self.screen_to_pixel(s_pos))		
