import random

import pygame
from pygame.locals import *

from helper import *
import story
import display
import mob
import menu
import terrain

class Map:
	def __init__(self):
		self.t_xsize = TILE_MAP_WIDTH
		self.t_ysize = TILE_MAP_HEIGHT
		self.tilex = TILE_PIXEL_WIDTH
		self.tiley = TILE_PIXEL_HEIGHT
		self.t_goal = (2, 2)
		self.t_origin = (self.t_xsize - 1, self.t_ysize - 1)
		self.p_xsize = (self.t_xsize + self.t_ysize) * self.tilex / 2
		self.p_ysize = (self.t_xsize + self.t_ysize) * self.tiley / 2
		self.tiles = [[1 for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		for x in range(4):
			for y in range(4):
				self.tiles[x][y] = 3
		for x in range(3):
			for y in range(3):
				self.tiles[x][y] = 2
		for x in range(4):
			for y in range(4):
				self.tiles[self.t_xsize - 1 - x][self.t_ysize - 1 - y] = 5
		self.terrain = [[None for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		self.weights = [[1 for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		self.seek = [[9001 for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		self.escape = [[9001 for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		self.terrains = []
		self.mobs = []
		self.invaders = []
		self.projectiles = []
		self.invader_count = 0
		self.invaders_left = 0
		self.invader_rate = 100
		self.rate_scale = 1.0
		self.invader_delay = 60
		self.per_invade = 1
		self.dragon = False
		self.calculate_pathing()
		
		if DEBUG:
			self.invader_count = 1
			self.invaders_left = 1
	
	def update(self, tick):
		#if (tick % self.invader_rate) == 0:
		self.invader_delay -= 1
		if self.invader_delay <= 0 and self.invader_count > 0:
			self.invader_rate = max(2, int(self.invader_rate * self.rate_scale))
			self.send_invader()
			self.invader_delay = self.invader_rate
		
		for x in self.mobs + self.terrains + self.projectiles:
			x.update(tick)
	
	def check_story(self):
		pass

	def heighway_summoned(self):
		pass
	
	def heal_mobs(self):
		for mob in self.mobs:
			if mob.damage > 0:
				mob.damage -= min(mob.damage, mob.vitality / 50)
	
	def add_mob(self, m, t_pos):
		t_pos = tup2rect(t_pos)
		if isinstance(m, mob.ChaosWyrm):
			self.heighway_summoned()
			self.dragon = True
		display.get_screen().add_sprite(m.sprite, t_pos)
		m.set_position(t_pos)
		self.mobs.append(m)
	
	def del_mob(self, mob):
		display.get_screen().del_sprite(mob.sprite)
		self.mobs.remove(mob)
		if mob.enemy:
			self.invaders_left -= 1
	
	def add_projectile(self, proj, t_pos):
		display.get_screen().add_sprite(proj.sprite, t_pos)
		proj.set_position(t_pos)
		self.projectiles.append(proj)
	
	def del_projectile(self, proj):
		display.get_screen().del_sprite(proj.sprite)
		self.projectiles.remove(proj)
	
	def add_terrain(self, terrain, t_pos):
		t_pos = tup2rect(t_pos)
		self.terrain[t_pos.left][t_pos.top] = terrain
		self.terrains.append(terrain)
		display.get_screen().add_sprite(terrain.sprite, t_pos)
		terrain.set_position(t_pos)
		if terrain.path_weight > 0:
			self.weights[t_pos.left][t_pos.top] = terrain.path_weight
			self.calculate_pathing()
	
	def del_terrain(self, terrain):
		t_pos = terrain.position
		self.terrains.remove(terrain)
		self.terrain[t_pos.left][t_pos.top] = None
		self.weights[t_pos.left][t_pos.top] = 1
		display.get_screen().del_sprite(terrain.sprite)
		self.calculate_pathing()
	
	def send_invader(self):
		for x in range(self.per_invade):
			type = random.choice(self.invaders)
			self.add_mob(type(), self.t_origin)
			self.invader_count -= 1
	
	def can_summon_at(self, t_pos):
		if t_pos == None or not self.valid_tile(t_pos):
			return False
		if self.tiles[t_pos[0]][t_pos[1]] != 1 and not DEBUG:
			return False
		for m in self.mobs:
			if m.position == t_pos:
				return False
		return self.terrain[t_pos[0]][t_pos[1]] == None

	def summon(self, summon, t_pos):
		if isinstance(summon, mob.Mob):
			self.add_mob(summon, t_pos)
		elif isinstance(summon, terrain.Terrain):
			self.add_terrain(summon, t_pos)
	
	def rand_from(self, t_pos):
		if isinstance(t_pos, pygame.Rect):
			t_pos = t_pos.topleft
		return random.choice(self.adjacent_nodes(t_pos))
	
	def seek_from(self, t_pos):
		if isinstance(t_pos, pygame.Rect):
			t_pos = t_pos.topleft
		score = 9001
		nodes = [t_pos]
		for n in self.adjacent_nodes(t_pos):
			if self.seek_score(n) < score:
				score = self.seek_score(n)
				nodes = [n]
			elif self.seek_score(n) == score:
				nodes.append(n)
		node = random.choice(nodes)
		return tup2rect(node)
	
	def escape_from(self, t_pos):
		if isinstance(t_pos, pygame.Rect):
			t_pos = t_pos.topleft
		score = 9001
		nodes = [t_pos]
		for n in self.adjacent_nodes(t_pos):
			if self.escape_score(n) < score:
				score = self.escape_score(n)
				nodes = [n]
			elif self.escape_score(n) == score:
				nodes.append(n)
		node = random.choice(nodes)
		return tup2rect(node)
	
	# this is sort of like A* pathfinding
	def move_toward(self, t_from, t_to, weight_limit=15):
		tick = pygame.time.get_ticks()
		start_node = t_from.topleft
		end_node = t_to.topleft
		open_nodes = [start_node]
		closed_nodes = []
		grid = [[None for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		grid[start_node[0]][start_node[1]] = (0, None)
		while open_nodes != [] and not end_node in open_nodes:
			guessing = 9001
			score = 9001
			info = None
			for n in open_nodes:
				manhattan = abs(n[0] - end_node[0]) + abs(n[1] - end_node[1])
				info = grid[n[0]][n[1]]
				if (info[0] + manhattan) <= guessing:
					guessing = info[0] + manhattan
					score = info[0]
					node = n
			if score <= weight_limit:
				adjacents = self.adjacent_nodes(node)
				open_nodes.remove(node)
				closed_nodes.append(node)
				for n in adjacents:
					new_score = score + self.weights[n[0]][n[1]]
					for m in self.mobs:
						if m.position.topleft == n:
							new_score += 1000
					if not (n in open_nodes) and not (n in closed_nodes):
						open_nodes.append(n)
						grid[n[0]][n[1]] = (new_score, node)
					elif new_score < grid[n[0]][n[1]][0]:
						grid[n[0]][n[1]] = (new_score, node)
			else:
				# giving up, too far
				return None
		
		node = end_node
		while grid[node[0]][node[1]][1] != start_node:
			node = grid[node[0]][node[1]][1]
		
		return node
		
	
	def calculate_pathing(self):
		self.seek = [[9001 for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		self.escape = [[9001 for y in range(self.t_ysize)] for x in range(self.t_xsize)]
		self.seek[self.t_goal[0]][self.t_goal[1]] = 0
		nodes = self.adjacent_nodes(self.t_goal)
		while (nodes != []):
			node = nodes[0]
			adjacents = self.adjacent_nodes(nodes[0])
			cur = self.seek[node[0]][node[1]]
			new = self.weights[node[0]][node[1]] + min([self.seek_score(a) for a in adjacents])
			if new < cur:
				self.seek[node[0]][node[1]] = new
				nodes = nodes[1:] + adjacents
			else:
				nodes = nodes[1:]
		
		self.escape[self.t_origin[0]][self.t_origin[1]] = 0
		nodes = self.adjacent_nodes(self.t_origin)
		while (nodes != []):
			node = nodes[0]
			adjacents = self.adjacent_nodes(nodes[0])
			cur = self.escape[node[0]][node[1]]
			new = self.weights[node[0]][node[1]] + min([self.escape_score(a) for a in adjacents])
			if new < cur:
				self.escape[node[0]][node[1]] = new
				nodes = nodes[1:] + adjacents
			else:
				nodes = nodes[1:]
		#print self.escape
	
	def seek_score(self, node):
		return self.seek[node[0]][node[1]]
	
	def escape_score(self, node):
		return self.escape[node[0]][node[1]]
	
	def adjacent_nodes(self, node):
		n = []
		if node[0] > 0:
			n.append((node[0] - 1, node[1]))
		if node[1] > 0:
			n.append((node[0], node[1] - 1))
		if node[0] < self.t_xsize - 1:
			n.append((node[0] + 1, node[1]))
		if node[1] < self.t_ysize - 1:
			n.append((node[0], node[1] + 1))
		return n
	
	
	def tile_to_pixel(self, tile_pos):
		tile_pos = tup2rect(tile_pos)
		x = (self.p_xsize - self.tilex + (tile_pos.left * self.tilex) - (tile_pos.top * self.tilex)) / 2
		y = (tile_pos.left + tile_pos.top) * self.tiley / 2
		return pygame.Rect(x, y, 0, 0)
	
	def pixel_to_tile(self, pixel_pos):
		pixel_pos = tup2rect(pixel_pos)
		x = (pixel_pos.top + ((pixel_pos.left - (self.p_xsize / 2)) / 2)) / (self.tilex / 2)
		y = (pixel_pos.top - ((pixel_pos.left - (self.p_xsize / 2)) / 2)) / self.tiley
		return pygame.Rect(x, y, 0, 0)
	
	def valid_tile(self, t_pos):
		t_pos = tup2rect(t_pos)
		return pygame.Rect(0, 0, self.t_xsize, self.t_ysize).contains(t_pos)
	
	def render(self):
		for x in range(self.t_xsize):
			for y in range(self.t_ysize):
				display.get_screen().draw_tile(self.tiles[x][y], self.tile_to_pixel((x, y)))

	def render_tile(self, t_pos):
		display.get_screen().draw_tile(self.tiles[t_pos.left][t_pos.top], self.tile_to_pixel((t_pos.left, t_pos.top)))

	def you_fail_it(self):
		story.show_dialogue(story.Conversation.Loss)
		self.lost = True

class FirstMap(Map):
	def __init__(self):
		Map.__init__(self)
		self.story_shown = False
		self.invaders = [mob.Brat]
		self.invader_count = 5
		self.invaders_left = 5
		self.invader_rate = 120
		self.lost = False
		self.won = False
		menu.summon_actions = [mob.Slime()]

	def update(self, tick):
		if self.invaders_left == 0:
			self.won = True
		Map.update(self, tick)

	def check_story(self):
		if not self.story_shown:
			story.show_dialogue(story.Conversation.L1)
			self.story_shown = True


class SecondMap(Map):
	def __init__(self):
		Map.__init__(self)
		self.story_shown = False
		self.invaders = [mob.Brat]
		self.invader_count = 15
		self.invaders_left = 15
		self.invader_rate = 100
		self.invader_delay = 80
		self.lost = False
		self.won = False
		menu.summon_actions = [mob.Slime(), terrain.BrickWall(), terrain.HedgeWall(),
			terrain.PitTrap()]

	def update(self, tick):
		if self.invaders_left == 0:
			self.won = True
		Map.update(self, tick)

	def check_story(self):
		if not self.story_shown:
			story.show_dialogue(story.Conversation.L2)
			self.story_shown = True


class ThirdMap(Map):
	def __init__(self):
		Map.__init__(self)
		self.story_shown = False
		self.invaders = [mob.Girl, mob.Brat]
		self.invader_count = 20
		self.invaders_left = 20
		self.per_invade = 1
		self.invader_rate = 180
		self.rate_scale = 0.99
		self.invader_delay = 100
		self.lost = False
		self.won = False
		menu.summon_actions = [mob.Slime(), mob.RockGolem(),
			terrain.BrickWall(), terrain.HedgeWall(),
			terrain.PitTrap(), terrain.FlameTower()]

	def update(self, tick):
		if self.invaders_left == 0:
			self.won = True
		Map.update(self, tick)

	def check_story(self):
		if not self.story_shown:
			story.show_dialogue(story.Conversation.L3)
			self.story_shown = True

class FourthMap(Map):
	def __init__(self):
		Map.__init__(self)
		self.story_shown = False
		self.invaders = [mob.Dog, mob.Brat, mob.Brat, mob.Girl]
		self.invader_count = 30
		self.invaders_left = 30
		self.invader_rate = 100
		self.per_invade = 2
		self.rate_scale = 0.97
		self.lost = False
		self.won = False
		menu.summon_actions = [mob.Slime(), mob.RockGolem(),
			terrain.BrickWall(), terrain.HedgeWall(),
			terrain.PitTrap(), terrain.ManaTrap(), terrain.FlameTower()]

	def update(self, tick):
		if self.invaders_left == 0:
			self.won = True
		Map.update(self, tick)

	def check_story(self):
		if not self.story_shown:
			story.show_dialogue(story.Conversation.L4)
			self.story_shown = True


class FifthMap(Map):
	def __init__(self):
		Map.__init__(self)
		self.story_shown = False
		self.waves = [[mob.Brat], [mob.Dog]]
		self.invaders = [mob.Brat]
		self.invader_count = 45
		self.invaders_left = 45
		self.invader_rate = 180
		self.lost = False
		self.won = False
		menu.summon_actions = menu.all_summon_actions

	def heighway_summoned(self):
		story.show_dialogue(story.Conversation.Dragon)

	def update(self, tick):
		if self.invader_count <= 40:
			self.invader_rate = 25
			self.per_invade = 2
			self.rate_scale = 0.96
			self.invaders = self.waves[1]
		if self.invaders_left == 0:
			self.won = True
		Map.update(self, tick)

	def check_story(self):
		if not self.story_shown:
			story.show_dialogue(story.Conversation.L5)
			self.story_shown = True

class SixthMap(Map):
	def __init__(self):
		Map.__init__(self)
		self.story_shown = False
		self.invaders = [mob.Dog, mob.Girl, mob.Brat]
		self.invader_count = 60
		self.invaders_left = 60
		self.invader_rate = 140
		self.per_invade = 2
		self.lost = False
		self.won = False
		self.win_shown = False
		self.rate_scale = 0.95
		menu.summon_actions = menu.all_summon_actions

	def update(self, tick):
		if self.invaders_left == 0:
			self.won = True
			if not self.win_shown:
				self.win_shown = True
				story.show_dialogue(story.Conversation.Ending)
		Map.update(self, tick)

	def check_story(self):
		if not self.story_shown:
			story.show_dialogue(story.Conversation.L6)
			self.story_shown = True





